import random

def heapify(nums, heap_size, root_index):
    # Assume the index of the largest element is the root index
    largest = root_index
    left_child = (2 * root_index) + 1
    right_child = (2 * root_index) + 2

    # If the left child of the root is a valid index, and the element is greater
    # than the current largest element, then update the largest element
    if left_child < heap_size and nums[left_child] > nums[largest]:
        largest = left_child

    # Do the same for the right child of the root
    if right_child < heap_size and nums[right_child] > nums[largest]:
        largest = right_child

    # If the largest element is no longer the root element, swap them
    if largest != root_index:
        nums[root_index], nums[largest] = nums[largest], nums[root_index]
        # Heapify the new root element to ensure it's the largest

        # využití tzv. rekurze - budeme probírat v dalším cvičení
        heapify(nums, heap_size, largest)


def heap_sort(nums):
    n = len(nums)

    # Create a Max Heap from the list
    # The 2nd argument of range means we stop at the element before -1 i.e.
    # the first element of the list.
    # The 3rd argument of range means we iterate backwards, reducing the count
    # of i by 1
    for i in range(n, -1, -1):
        heapify(nums, n, i)

    # Move the root of the max heap to the end of
    for i in range(n - 1, 0, -1):
        nums[i], nums[0] = nums[0], nums[i]
        heapify(nums, i, 0)
    
    return nums


def merge_sort(unsorted_array):
    if len(unsorted_array) > 1:
        mid = len(unsorted_array) // 2  # Finding the mid of the array
        left = unsorted_array[:mid]  # Dividing the array elements
        right = unsorted_array[mid:]  # into 2 halves

        # využití tzv. rekurze - budeme probírat v dalším cvičení
        merge_sort(left)
        merge_sort(right)

        i = j = k = 0

        #  data to temp arrays L[] and R[]
        while i < len(left) and j < len(right):
            if left[i] < right[j]:
                unsorted_array[k] = left[i]
                i += 1
            else:
                unsorted_array[k] = right[j]
                j += 1
            k += 1

        # Checking if any element was left
        while i < len(left):
            unsorted_array[k] = left[i]
            i += 1
            k += 1

        while j < len(right):
            unsorted_array[k] = right[j]
            j += 1
            k += 1

    return unsorted_array


def bubble_sort(a):
    n = len(a)
    sorted = False
    
    while not sorted:
        sorted = True # Initialize the sorted flag by True
        for i in range(n - 1):
            if a[i] > a[i + 1]: # If a bad order occurs
                a[i], a[i + 1] = a[i + 1], a[i] # Swap the elements
                sorted = False # Mark the array as not sorted
                
    return a # nemusi byt uvedeno (razeni listu na miste)


def bubble_sort_2(a):
    """ sorts array a in-place in ascending order"""
    for i in range(len(a)-1,0,-1):
        # i=n-1..1. a[i+1:] is already sorted
        exchanged=False # exchanges in this iteration?
        for j in range(i):
            if a[j]>a[j+1]:
                a[j],a[j+1]=a[j+1],a[j] # exchange
                exchanged=True
        if not exchanged: break

    return a


def insertion_sort(a):
    """ sorts array a in-place in ascending order"""
    for i in range(1,len(a)): # a[0:i] is sorted
        val=a[i]
        j=i
        while j>0 and a[j-1]>val:
            a[j]=a[j-1]
            j-=1
            a[j]=val

    return a


def selection_sort(a):
    """ sorts array a in-place in ascending order"""
    for i in range(len(a)-1,0,-1):
        # find out what should go to a[i]
        max_pos=0 # index of the maximum
        for j in range(1,i+1):
            if a[j]>a[max_pos]:
                max_pos=j
            a[i],a[max_pos]=a[max_pos],a[i]

    return a


def get_unsorted(length=100):
    """
    Unsorted array of integers.
    
    Args:
        length (int): Number of elements in the returned array. *Default: 100.*
    
    Returns:
        list[int]: An array full of random integers ranging from -100 to 100 (100 not included).
    """
    return [random.randrange(-100, 100) for i in range(length)]


def test_sort(f=bubble_sort, repeats=100, length=100):
    for j in range(repeats):
        a=[random.randrange(100000) for i in range(length)]
        
        if f.__name__ == 'sorted':
            a = f(a)
        else:
            f(a)

        for i in range(length-1):

            assert(a[i]<=a[i+1])
        
            # assert <podmínka> vrátí chybu v případě, že je <podmínka> == False
            # podobně jako:
            # if not <podmínka>: raise AssertionError()
        
    print(f.__name__,"sort test passed")